@@ -9,6 +9,8 @@
 #' @param env.nback Number of points to draw from environment space for environment space discrimination metrics.
 #' @param rts.reps The number of replicates to do for a Raes and ter Steege-style test of significance
 #' @param bg.source Source for drawing background points.  If "points", it just uses the background points that are already in the species object.  If "range", it uses the range raster.  If "env", it draws points at randome from the entire study area outlined by the first environmental layer.
+#' @param verbose Controls printing of various messages progress reports.  Defaults to FALSE.
+#' @param clamp When set to TRUE, clamps the environmental layers so that predictions made outside the min/max of the training data for each predictor are set to the value for the min/max for that predictor. Prevents the model from extrapolating beyond the min/max bounds of the predictor space the model was trained in, although there could still be projections outside the multivariate training space if predictors are strongly correlated.
 #' @param ... Arguments to be passed to domain()
 #'
 #' @return An enmtools model object containing species name, model formula (if any), model object, suitability raster, marginal response plots, and any evaluation objects that were created.
@@ -21,11 +23,11 @@
 #' }
 
 
-enmtools.dm <- function(species, env = NA, test.prop = 0, report = NULL, nback = 1000, env.nback = 10000, overwrite = FALSE, rts.reps = 0, bg.source = "default", ...){
+enmtools.dm <- function(species, env = NA, test.prop = 0, report = NULL, nback = 1000, env.nback = 10000, overwrite = FALSE, rts.reps = 0, bg.source = "default", verbose = FALSE, clamp = TRUE, ...){
 
   notes <- NULL
 
-  species <- check.bg(species, env, nback)
+  species <- check.bg(species, env, nback, verbose = verbose)
 
   dm.precheck(species, env)
 
@@ -70,6 +72,20 @@ enmtools.dm <- function(species, env = NA, test.prop = 0, report = NULL, nback =
 
   this.dm <- dismo::domain(env, species$presence.points[,1:2])
 
+  suitability <- predict(env, this.dm, type = "response")
+
+  # Clamping and getting a diff layer
+  clamping.strength <- NA
+  if(clamp == TRUE){
+    # Adding env (skipped for BC otherwise)
+    this.df <- as.data.frame(extract(env, species$presence.points))
+
+    env <- clamp.env(this.df, env)
+    clamped.suitability <- predict(env, this.dm, type = "response")
+    clamping.strength <- clamped.suitability - suitability
+    suitability <- clamped.suitability
+  }
+
   model.evaluation <- dismo::evaluate(species$presence.points[,1:2], species$background.points[,1:2],
                                this.dm, env)
   env.model.evaluation <- env.evaluate(species, this.dm, env, n.background = env.nback)
@@ -101,11 +117,11 @@ enmtools.dm <- function(species, env = NA, test.prop = 0, report = NULL, nback =
     }
   }
 
-  suitability <- predict(env, this.dm, type = "response")
-
   # Do Raes and ter Steege test for significance.  Turned off if eval == FALSE
   if(rts.reps > 0){
 
+    message("\nBuilding RTS replicate models...\n")
+
     # Die if we're not doing randomly withheld test data and RTS reps > 0
     if(!is.numeric(test.prop)){
       stop(paste("RTS test can only be conducted with randomly withheld data, and test.prop is set to", test.prop))
@@ -118,9 +134,19 @@ enmtools.dm <- function(species, env = NA, test.prop = 0, report = NULL, nback =
     rts.env.training <- c()
     rts.env.test <- c()
 
+    if (requireNamespace("progress", quietly = TRUE)) {
+      pb <- progress::progress_bar$new(
+        format = " [:bar] :percent eta: :eta",
+        total = rts.reps, clear = FALSE, width= 60)
+    }
+
     for(i in 1:rts.reps){
 
-      message(paste("Replicate", i, "of", rts.reps))
+      if (requireNamespace("progress", quietly = TRUE)) {
+        pb$tick()
+      }
+
+      if(verbose == TRUE){message(paste("Replicate", i, "of", rts.reps))}
 
       # Repeating analysis with scrambled pa points and then evaluating models
       rep.species <- species
@@ -244,6 +270,7 @@ enmtools.dm <- function(species, env = NA, test.prop = 0, report = NULL, nback =
                  env.test.evaluation = env.test.evaluation,
                  rts.test = rts.test,
                  suitability = suitability,
+                 clamping.strength = clamping.strength,
                  call = sys.call(),
                  notes = notes)
 
@@ -326,7 +353,11 @@ plot.enmtools.dm <- function(x, ...){
     geom_raster(aes_string(fill = "Suitability")) +
     scale_fill_viridis_c(option = "B", guide = guide_colourbar(title = "Suitability")) +
     coord_fixed() + theme_classic() +
-    geom_point(data = x$analysis.df, aes_string(y = "Latitude", x = "Longitude"),
+    geom_point(data = x$analysis.df[x$analysis.df$presence == 1 &
+                                      x$analysis.df$Longitude > extent(x$suitability)[1] &
+                                      x$analysis.df$Longitude < extent(x$suitability)[2] &
+                                      x$analysis.df$Latitude > extent(x$suitability)[3] &
+                                      x$analysis.df$Latitude < extent(x$suitability)[4],], aes_string(y = "Latitude", x = "Longitude"),
                pch = 21, fill = "white", color = "black", size = 2)
 
   if(!(all(is.na(x$test.data)))){
@@ -345,7 +376,7 @@ plot.enmtools.dm <- function(x, ...){
 
 
 # Predict method for models of class enmtools.dm
-predict.enmtools.dm <- function(object, env, maxpts = 1000, ...){
+predict.enmtools.dm <- function(object, env, maxpts = 1000, clamp = TRUE, ...){
 
   # The domain predict function doesn't like having layers in the stack that aren't in
   # the presence df
@@ -353,6 +384,19 @@ predict.enmtools.dm <- function(object, env, maxpts = 1000, ...){
 
   # Make a plot of habitat suitability in the new region
   suitability <- raster::predict(env, object$model)
+
+  # Clamping and getting a diff layer
+  clamping.strength <- NA
+  if(clamp == TRUE){
+    # Adding env (skipped for BC otherwise)
+    this.df <- as.data.frame(extract(env, object$analysis.df))
+
+    env <- clamp.env(this.df, env)
+    clamped.suitability <- raster::predict(env, object$model)
+    clamping.strength <- clamped.suitability - suitability
+    suitability <- clamped.suitability
+  }
+
   suit.points <- data.frame(rasterToPoints(suitability))
   colnames(suit.points) <- c("Longitude", "Latitude", "Suitability")
 
@@ -361,6 +405,14 @@ predict.enmtools.dm <- function(object, env, maxpts = 1000, ...){
     scale_fill_viridis_c(option = "B", guide = guide_colourbar(title = "Suitability")) +
     coord_fixed() + theme_classic()
 
+  clamp.points <- data.frame(rasterToPoints(clamping.strength))
+  colnames(clamp.points) <- c("Longitude", "Latitude", "Clamping")
+
+  clamp.plot <- ggplot(data = clamp.points,  aes_string(y = "Latitude", x = "Longitude")) +
+    geom_raster(aes_string(fill = "Clamping")) +
+    scale_fill_viridis_c(option = "B", guide = guide_colourbar(title = "Suitability")) +
+    coord_fixed() + theme_classic()
+
   if(!is.na(object$species.name)){
     title <- paste("Domain model projection for", object$species.name)
     suit.plot <- suit.plot + ggtitle(title) + theme(plot.title = element_text(hjust = 0.5))
@@ -368,8 +420,10 @@ predict.enmtools.dm <- function(object, env, maxpts = 1000, ...){
 
   this.threespace = threespace.plot(object, env, maxpts)
 
-  output <- list(suitability = suit.plot,
-                 raster = suitability,
+  output <- list(suitability.plot = suit.plot,
+                 suitability = suitability,
+                 clamping.strength = clamping.strength,
+                 clamp.plot = clamp.plot,
                  threespace.plot = this.threespace)
   return(output)
 }
