@@ -12,6 +12,8 @@
 #' @param rts.reps The number of replicates to do for a Raes and ter Steege-style test of significance
 #' @param weights If this is set to "equal", presences and background data will be assigned weights so that the sum of all presence points weights equals the sum of all background point weights.  Otherwise, weights are not provided to the model.
 #' @param bg.source Source for drawing background points.  If "points", it just uses the background points that are already in the species object.  If "range", it uses the range raster.  If "env", it draws points at randome from the entire study area outlined by the first environmental layer.
+#' @param verbose Controls printing of various messages progress reports.  Defaults to FALSE.
+#' @param clamp When set to TRUE, clamps the environmental layers so that predictions made outside the min/max of the training data for each predictor are set to the value for the min/max for that predictor. Prevents the model from extrapolating beyond the min/max bounds of the predictor space the model was trained in, although there could still be projections outside the multivariate training space if predictors are strongly correlated.
 #' @param ... Arguments to be passed to glm()
 #'
 #' @return An enmtools model object containing species name, model formula (if any), model object, suitability raster, marginal response plots, and any evaluation objects that were created.
@@ -23,11 +25,11 @@
 
 
 
-enmtools.glm <- function(species, env, f = NULL, test.prop = 0, eval = TRUE, nback = 1000, env.nback = 10000, report = NULL, overwrite = FALSE, rts.reps = 0, weights = "equal", bg.source = "default", ...){
+enmtools.glm <- function(species, env, f = NULL, test.prop = 0, eval = TRUE, nback = 1000, env.nback = 10000, report = NULL, overwrite = FALSE, rts.reps = 0, weights = "equal", bg.source = "default",  verbose = FALSE, clamp = TRUE, ...){
 
   notes <- NULL
 
-  species <- check.bg(species, env, nback = nback, bg.source = bg.source)
+  species <- check.bg(species, env, nback = nback, bg.source = bg.source, verbose = verbose)
 
   # Builds a default formula using all env
   if(is.null(f)){
@@ -75,7 +77,7 @@ enmtools.glm <- function(species, env, f = NULL, test.prop = 0, eval = TRUE, nba
   #                          analysis.df[analysis.df$presence == 0,], 2)
 
   ### Add env data
-  species <- add.env(species, env)
+  species <- add.env(species, env, verbose = verbose)
 
   # Recast this formula so that the response variable is named "presence"
   # regardless of what was passed.
@@ -101,6 +103,15 @@ enmtools.glm <- function(species, env, f = NULL, test.prop = 0, eval = TRUE, nba
 
   suitability <- predict(env, this.glm, type = "response")
 
+  # Clamping and getting a diff layer
+  clamping.strength <- NA
+  if(clamp == TRUE){
+    env <- clamp.env(analysis.df, env)
+    clamped.suitability <- predict(env, this.glm, type = "response")
+    clamping.strength <- clamped.suitability - suitability
+    suitability <- clamped.suitability
+  }
+
   if(eval == TRUE){
 
     # This is a very weird hack that has to be done because dismo's evaluate function
@@ -147,6 +158,8 @@ enmtools.glm <- function(species, env, f = NULL, test.prop = 0, eval = TRUE, nba
     # Do Raes and ter Steege test for significance.  Turned off if eval == FALSE
     if(rts.reps > 0 & eval == TRUE){
 
+      message("\nBuilding RTS replicate models...\n")
+
       # Die if we're not doing randomly withheld test data and RTS reps > 0
       if(!is.numeric(test.prop)){
         stop(paste("RTS test can only be conducted with randomly withheld data, and test.prop is set to", test.prop))
@@ -159,9 +172,19 @@ enmtools.glm <- function(species, env, f = NULL, test.prop = 0, eval = TRUE, nba
       rts.env.training <- c()
       rts.env.test <- c()
 
+      if (requireNamespace("progress", quietly = TRUE)) {
+        pb <- progress::progress_bar$new(
+          format = " [:bar] :percent eta: :eta",
+          total = rts.reps, clear = FALSE, width= 60)
+      }
+
       for(i in 1:rts.reps){
 
-        message(paste("Replicate", i, "of", rts.reps))
+        if (requireNamespace("progress", quietly = TRUE)) {
+          pb$tick()
+        }
+
+        if(verbose == TRUE){message(paste("Replicate", i, "of", rts.reps))}
 
         # Repeating analysis with scrambled pa points and then evaluating models
         rep.species <- species
@@ -184,7 +207,7 @@ enmtools.glm <- function(species, env, f = NULL, test.prop = 0, eval = TRUE, nba
         # Everything else goes back to the background
         rep.species$background.points <- allpoints
 
-        rep.species <- add.env(rep.species, env, verbose = FALSE)
+        rep.species <- add.env(rep.species, env, verbose = verbose)
 
         rts.df <- rbind(rep.species$presence.points, rep.species$background.points)
         rts.df$presence <- c(rep(1, nrow(rep.species$presence.points)), rep(0, nrow(rep.species$background.points)))
@@ -292,6 +315,7 @@ enmtools.glm <- function(species, env, f = NULL, test.prop = 0, eval = TRUE, nba
                  env.test.evaluation = env.test.evaluation,
                  rts.test = rts.test,
                  suitability = suitability,
+                 clamping.strength = clamping.strength,
                  call = sys.call(),
                  notes = notes)
 
@@ -402,10 +426,20 @@ plot.enmtools.glm <- function(x, ...){
 
 
 # Predict method for models of class enmtools.glm
-predict.enmtools.glm <- function(object, env, maxpts = 1000, ...){
+predict.enmtools.glm <- function(object, env, maxpts = 1000, clamp = TRUE, ...){
 
   # Make a plot of habitat suitability in the new region
   suitability <- raster::predict(env, object$model, type = "response")
+
+  # Clamping and getting a diff layer
+  clamping.strength <- NA
+  if(clamp == TRUE){
+    env <- clamp.env(object$analysis.df, env)
+    clamped.suitability <- raster::predict(env, object$model, type = "response")
+    clamping.strength <- clamped.suitability - suitability
+    suitability <- clamped.suitability
+  }
+
   suit.points <- data.frame(rasterToPoints(suitability))
   colnames(suit.points) <- c("Longitude", "Latitude", "Suitability")
 
@@ -414,6 +448,14 @@ predict.enmtools.glm <- function(object, env, maxpts = 1000, ...){
     scale_fill_viridis_c(option = "B", guide = guide_colourbar(title = "Suitability")) +
     coord_fixed() + theme_classic()
 
+  clamp.points <- data.frame(rasterToPoints(clamping.strength))
+  colnames(clamp.points) <- c("Longitude", "Latitude", "Clamping")
+
+  clamp.plot <- ggplot(data = clamp.points,  aes_string(y = "Latitude", x = "Longitude")) +
+    geom_raster(aes_string(fill = "Clamping")) +
+    scale_fill_viridis_c(option = "B", guide = guide_colourbar(title = "Suitability")) +
+    coord_fixed() + theme_classic()
+
   if(!is.na(object$species.name)){
     title <- paste("GLM model projection for", object$species.name)
     suit.plot <- suit.plot + ggtitle(title) + theme(plot.title = element_text(hjust = 0.5))
@@ -421,8 +463,10 @@ predict.enmtools.glm <- function(object, env, maxpts = 1000, ...){
 
   this.threespace = threespace.plot(object, env, maxpts)
 
-  output <- list(suitability = suit.plot,
-                 raster = suitability,
+  output <- list(suitability.plot = suit.plot,
+                 clamping.strength = clamping.strength,
+                 suitability = suitability,
+                 clamp.plot = clamp.plot,
                  threespace.plot = this.threespace)
   return(output)
 }
